/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#pragma once

#define LOG_SENDER_FRAME_DROPS false

#include "ExpCamera.h"
#include "ExpPipelineTypes.h"
#include "MixCastLightComponent.h"
#include "Transmission/Unreal-Interop/SharedQueue/SharedQueueProducer.hpp"
#include "Transmission/Unreal-Interop/SharedTextures/OwnedUnrealSharedTexture_D3D11.h"
#include "Data/SharedTexIds.h"
#include "Data/SdkData.h"
#include "MixCastLog.h"
#include "Utility/StringUtils.hpp"
#include "Utility/TimeUtils.hpp"

#include "CoreMinimal.h"
#include "Engine/World.h"
#include "GameFramework/WorldSettings.h"
#include "Kismet/GameplayStatics.h"
#include "Components/LightComponent.h"
#include "Components/PointLightComponent.h"
#include "Components/DirectionalLightComponent.h"
#include "Components/SpotLightComponent.h"

class MIXCAST_API FExpFrameSender
{
public:
	const int QueueCapacity = 10;

private:
	AExpCamera* cam = NULL;
	FDelegateHandle frameDelegate;

	SharedQueueProducer<ExpFrame>* renderedExpFrames = NULL;
	OwnedUnrealSharedTexture** queueTextures;

	bool firstWriteCompleted = false;
	bool consumerMissedLastFrame = false;

	void PopulateLights(ExpLight lights[])
	{
		FTransform trackingRootInv = AMixCastSdkActor::GetWorldTransform().Inverse();
		float worldToMeters = cam->GetWorld()->GetWorldSettings()->WorldToMeters;

		int listIndex = 0;
		int arrIndex = 0;
		while (arrIndex < ExpFrame::MaxLights)
		{
			if (listIndex >= UMixCastLightComponent::Active.Num())
			{
				lights[arrIndex++].type = ExpLightType::None;
				continue;
			}

			ULightComponent* light = UMixCastLightComponent::Active[listIndex++]->GetSource();				
			if (light->Intensity <= 0)
				continue;

			lights[arrIndex].color = FVector(light->LightColor.R, light->LightColor.G, light->LightColor.B) * light->Intensity;

			lights[arrIndex].pos = MathUtil::UnrealPositionToMixCast(trackingRootInv.TransformPosition(light->GetComponentLocation()), worldToMeters);
			lights[arrIndex].forward = MathUtil::UnrealVectorToMixCast(trackingRootInv.TransformVector(light->GetForwardVector()));
			lights[arrIndex].forward.Normalize();

			lights[arrIndex].range = 0;
			lights[arrIndex].angle = 0;

			UClass* lightClass = light->GetClass();
			if (lightClass == UPointLightComponent::StaticClass()) {
				lights[arrIndex].type = ExpLightType::Point;
				lights[arrIndex].range = ((UPointLightComponent*)light)->AttenuationRadius / worldToMeters;
			}
			else if (lightClass == UDirectionalLightComponent::StaticClass()) {
				lights[arrIndex].type = ExpLightType::Directional;
			}
			else if (lightClass == USpotLightComponent::StaticClass()) {
				lights[arrIndex].type = ExpLightType::Spot;
				lights[arrIndex].range = ((USpotLightComponent*)light)->AttenuationRadius / worldToMeters;
				lights[arrIndex].angle = FMath::DegreesToRadians(((USpotLightComponent*)light)->OuterConeAngle);
			}
			else
				continue;

			arrIndex++;
		}
	}

	void CreateQueue()
	{
		std::wstring srcId = StringUtils::ToStdWString(SharedTexIds::Cameras::RenderedExpFrames::Get(cam->Identifier));
		renderedExpFrames = new SharedQueueProducer<ExpFrame>(srcId.c_str(), QueueCapacity);
	}
	void DestroyQueue()
	{
		if (renderedExpFrames != NULL)
		{
			delete renderedExpFrames;
			renderedExpFrames = NULL;
		}
	}
	void CreateTextureList()
	{
		queueTextures = new OwnedUnrealSharedTexture * [QueueCapacity];
		for (int i = 0; i < QueueCapacity; i++)
			queueTextures[i] = NULL;
	}
	void DestroyTextureList()
	{
		if (queueTextures != NULL)
		{
			for (int i = 0; i < QueueCapacity; i++) {
				if (queueTextures[i] != NULL) {
					delete queueTextures[i];
					queueTextures[i] = NULL;
				}
			}
			delete[] queueTextures;
			queueTextures = NULL;
		}
	}

public:
	FExpFrameSender(AExpCamera* cam)
	{
		this->cam = cam;
		
		CreateQueue();
		CreateTextureList();

		frameDelegate = AExpCamera::OnFrameEnded().AddRaw(this, &FExpFrameSender::HandleFrameUpdated);
	}
	~FExpFrameSender()
	{
		AExpCamera::OnFrameEnded().Remove(frameDelegate);
		
		DestroyQueue();
		DestroyTextureList();
	}

	void HandleFrameUpdated(AExpCamera* curCam)
	{
		if (curCam != cam)
			return;

		renderedExpFrames->EnqueueFenceUpdates();

		UMixCastFrameCanvas* layersCanvas = cam->GetLayersCanvas();
		if (!layersCanvas->IsValidLowLevel() || !layersCanvas->Resource->TextureRHI || !layersCanvas->Resource->TextureRHI->GetNativeResource())
			return;

		ExpFrame renderedFrame = cam->GetLatestFrameInfo();

		if (!renderedExpFrames->WaitUntilNextEmptied(0))
		{
			if (!consumerMissedLastFrame && firstWriteCompleted && LOG_SENDER_FRAME_DROPS)
			{
				MIXCAST_WARNING("Compositing process started missing frames at time {%s}",
					*TimeUtils::GetSMPTE(TimeUtils::GetMidnightRelativeFrameIndex(renderedFrame.syncTime, renderedFrame.camFramerate), renderedFrame.camFramerate));
			}
			consumerMissedLastFrame = true;
			return;
		}

		//Copy layer tex
		OwnedUnrealSharedTexture* sharedTex = queueTextures[renderedExpFrames->GetCurrentSlot()];
		if (sharedTex == NULL)
		{
			sharedTex = OwnedUnrealSharedTexture::Create(layersCanvas);
			queueTextures[renderedExpFrames->GetCurrentSlot()] = sharedTex;
		}
		sharedTex->EnqueueWrite(layersCanvas);

		renderedFrame.layersTex = sharedTex->GetHandle();
		PopulateLights(renderedFrame.lights);

		renderedExpFrames->Write(renderedFrame, true);

		if (!firstWriteCompleted && LOG_SENDER_FRAME_DROPS)
		{
			firstWriteCompleted = true;
			consumerMissedLastFrame = false;

			MIXCAST_LOG("Compositing process started at time {%s}",
				*TimeUtils::GetSMPTE(TimeUtils::GetMidnightRelativeFrameIndex(renderedFrame.syncTime, renderedFrame.camFramerate), renderedFrame.camFramerate));
		}

		if (consumerMissedLastFrame && LOG_SENDER_FRAME_DROPS)
		{
			MIXCAST_LOG("Compositing process back in business at time {%s}",
				*TimeUtils::GetSMPTE(TimeUtils::GetMidnightRelativeFrameIndex(renderedFrame.syncTime, renderedFrame.camFramerate), renderedFrame.camFramerate));
		}

		consumerMissedLastFrame = false;
	}
};
