/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "PluginUpdater.h"

#include "MixCast.h"
#include "MixCastLog.h"
#include "Utility/ProjectUtils.hpp"
#include "Utility/StringUtils.hpp"

#include "Http.h"
#include "Misc/MessageDialog.h"

#define VERSION_URL "https://mixcast.me/downloads/info/unreal-sdk-version.txt"
#define DOWNLOAD_URL "https://mixcast.me/unreal-sdk"

#define UPDATES_CONFIG_SECTION TEXT("MixCast")
#define UPDATES_CONFIG_CHECK_KEY TEXT("NextUpdateCheckTime")
#define UNCHANGED_COOLDOWN 60 * 60
#define LATER_COOLDOWN 3 * 24 * 60 * 60

void FPluginUpdater::CheckForUpdates(bool headless)
{
#if UE_EDITOR
	static FString NeverCheckAgainValue("-");

	FString NextCheckStr;
	GConfig->GetString(UPDATES_CONFIG_SECTION, UPDATES_CONFIG_CHECK_KEY, NextCheckStr, GEditorIni);

	FDateTime Now = FDateTime::UtcNow();
	FString NowStr = Now.ToString();

	if (NextCheckStr.Compare(NeverCheckAgainValue) == 0)
		return;
	
	FDateTime NextCheck;
	if (FDateTime::Parse(NextCheckStr, NextCheck) && NextCheck > Now)
		return;

	auto HttpRequest = FHttpModule::Get().CreateRequest();
	HttpRequest->SetVerb("POST");
	HttpRequest->SetHeader("Content-Type", "application/x-www-form-urlencoded");
	HttpRequest->SetURL(TEXT(VERSION_URL));

	HttpRequest->OnProcessRequestComplete().BindLambda([headless](const FHttpRequestPtr Request, const FHttpResponsePtr Response, const bool bWasSuccessful)
	{
		if (!bWasSuccessful)
		{
			if (!headless) {
				FMessageDialog::Open(EAppMsgType::Ok, FText::FromString(TEXT("Couldn't contact the update server, try again later")));
			}
			return;
		}

		FDateTime NextCheck = FDateTime::UtcNow();

		const FString responseStr = Response->GetContentAsString();
		const FString versionStr = FMixCast::GetPluginVersion();
		if (StringUtils::IsVersionBLaterThanVersionA(versionStr, responseStr))
		{
			FString msg = FString::Format(TEXT("MixCast has a new version available ({0})! Do you want to download it now?"), {
				FStringFormatArg(responseStr),
			});
			EAppReturnType::Type downloadNowResponse = FMessageDialog::Open(EAppMsgType::YesNo, FText::FromString(msg));
			if (downloadNowResponse == EAppReturnType::Yes)
			{
				FPlatformProcess::LaunchURL(TEXT(DOWNLOAD_URL), NULL, NULL);

				NextCheck += FTimespan::FromSeconds(UNCHANGED_COOLDOWN);
			}
			else
			{
				EAppReturnType::Type remindResponse = FMessageDialog::Open(EAppMsgType::YesNo, FText::FromString(TEXT("OK! Do you still want to be notified of MixCast SDK updates in the future?")));
				if (remindResponse == EAppReturnType::Yes)
				{
					NextCheck += FTimespan::FromSeconds(LATER_COOLDOWN);
				}
				else
				{
					GConfig->SetString(UPDATES_CONFIG_SECTION, UPDATES_CONFIG_CHECK_KEY, *NeverCheckAgainValue, GEditorIni);
					return;
				}
			}
		}
		else if (StringUtils::IsVersionBLaterThanVersionA(responseStr, versionStr))
		{
			NextCheck += FTimespan::FromSeconds(UNCHANGED_COOLDOWN);

			if (!headless) {
				FString msg = FString::Format(TEXT("Your MixCast version ({0}) is ahead of the public release ({1})!"), { 
					FStringFormatArg(responseStr),
					FStringFormatArg(versionStr)
				});
				FMessageDialog::Open(EAppMsgType::Ok, FText::FromString(msg));
			}
		}
		else
		{
			NextCheck += FTimespan::FromSeconds(UNCHANGED_COOLDOWN);

			if (!headless) {
				FString msg = FString::Format(TEXT("MixCast is up-to-date at version {0}!"), { 
					FStringFormatArg(versionStr)
				});
				FMessageDialog::Open(EAppMsgType::Ok, FText::FromString(msg));
			}
		}

		FString NextCheckStr = NextCheck.ToString();

		GConfig->SetString(UPDATES_CONFIG_SECTION, UPDATES_CONFIG_CHECK_KEY, *NextCheckStr, GEditorIni);
	});
	HttpRequest->ProcessRequest();
#endif
}